package org.nakedobjects.nof.reflect.java.reflect;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import org.apache.log4j.Logger;
import org.nakedobjects.noa.adapter.NakedCollection;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.adapter.NakedObjectLoader;
import org.nakedobjects.noa.reflect.Consent;
import org.nakedobjects.noa.spec.NakedObjectSpecification;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.util.DebugString;
import org.nakedobjects.nof.reflect.peer.MemberIdentifier;
import org.nakedobjects.nof.reflect.peer.OneToManyPeer;
import org.nakedobjects.nof.reflect.peer.ReflectionException;


public abstract class AbstractJavaCollectionField extends JavaField implements OneToManyPeer, ExtensionHolderMutableFieldPeer {
    private final static Logger LOG = Logger.getLogger(AbstractJavaCollectionField.class);
    private final Method addMethod;
    private final Method clearMethod;
    private final Method removeMethod;
    private final Method addValidMethod;
    private final Method removeValidMethod;

    public AbstractJavaCollectionField(
            final MemberIdentifier name,
            final Class type,
            final DescriptiveMethods descriptiveMethods,
            final FieldMethods fieldMethods,
            final GeneralControlMethods controlMethods,
            final FieldFlags fieldFlags,
            final Method clearMethod,
            final MemberSessionMethods sessionMethods) {
        super(name, type, descriptiveMethods, fieldMethods, controlMethods, fieldFlags, sessionMethods);
        this.addMethod = remember(fieldMethods.getAssociateMethod());
        this.removeMethod = remember(fieldMethods.getDissociateMethod());
        this.addValidMethod = remember(controlMethods.getValidMethod1());
        this.removeValidMethod = remember(controlMethods.getValidMethod2());
        this.clearMethod = clearMethod;
    }

    public void addAssociation(final NakedObject inObject, final NakedObject associate) {
        LOG.debug("local set association " + getIdentifier() + " in " + inObject + " with " + associate);
        try {
            Object element = associate.getObject();
            if (addMethod == null) {
                add(getCollection(inObject), element);
                NakedObjectsContext.getObjectPersistor().objectChanged(inObject);
            } else {
                addMethod.invoke(inObject.getObject(), new Object[] { element });
            }
        } catch (IllegalArgumentException e) {
            throw new IllegalArgumentException("add method expects a " + getSpecification().getFullName() + " object; not a "
                    + associate.getClass().getName() + ", " + e.getMessage());
        } catch (InvocationTargetException e) {
            invocationException("Exception executing " + addMethod, e);
        } catch (IllegalAccessException ignore) {
            LOG.error("illegal access of " + addMethod, ignore);
            throw new RuntimeException(ignore.getMessage());
        }
    }

    protected Object getCollection(final NakedObject inObject) {
        try {
            return getMethod.invoke(inObject.getObject(), new Object[0]);
        } catch (InvocationTargetException e) {
            invocationException("Exception executing " + getMethod, e);
            return null; // the above method throws an exception so this statement is never reached
        } catch (IllegalAccessException ignore) {
            LOG.error("illegal access of " + addMethod, ignore);
            throw new RuntimeException(ignore.getMessage());
        }
    }

    protected abstract void add(final Object collection, Object element);

    public void debugData(final DebugString debug) {
        debug.appendln("Get method", getMethod);
        if (addMethod != null) {
            debug.appendln("Add method", addMethod);
        }
        if (removeMethod != null) {
            debug.appendln("Remove method", removeMethod);
        }
        if (clearMethod != null) {
            debug.appendln("Clear method", clearMethod);
        }
        if (addValidMethod != null) {
            debug.appendln("Valid add method", addValidMethod);
        }
        if (removeValidMethod != null) {
            debug.appendln("Valid remove method", removeValidMethod);
        }
        super.debugData(debug);
    }

    protected NakedCollection get(final NakedObject fromObject) {
        Object collection = getCollection(fromObject);
        NakedCollection adapter;
        if (collection == null) {
            return null;
        } else {
            NakedObjectLoader objectLoader = NakedObjectsContext.getObjectLoader();
            adapter = objectLoader.getAdapterForElseCreateAdapterForCollection(fromObject, getIdentifier().getName(), getSpecification(),
                    collection);
            if (adapter == null) {
                throw new ReflectionException("no adapter created");
            } else {
                return adapter;
            }
        }
    }

    public NakedCollection getAssociations(final NakedObject fromObject) {
        return (NakedCollection) get(fromObject);
    }

    /**
     * return the object type that the collection holds.
     */
    public NakedObjectSpecification getSpecification() {
        return type == null ? NakedObjectsContext.getReflector().loadSpecification(Object.class) : super.getSpecification();
    }

    public void initOneToManyAssociation(final NakedObject fromObject, final NakedObject[] elements) {
        Object collection = getCollection(fromObject);
        clear(collection);
        Object[] array = new Object[elements.length];
        for (int i = 0; i < elements.length; i++) {
            array[i] = elements[i].getObject();
        }
        init(collection, array);
    }

    protected abstract void init(Object collection, Object[] elements);

    public Consent isAddValid(final NakedObject container, final NakedObject element) {
        return executeConsent(addValidMethod, container, element, false);
    }

    public boolean isEmpty(final NakedObject fromObject) {
        Object collection = getCollection(fromObject);
        if (collection == null) {
            return true;
        } else {
            return isEmpty(collection);
        }
    }

    protected abstract boolean isEmpty(final Object collection);

    public Consent isRemoveValid(final NakedObject container, final NakedObject element) {
        return executeConsent(removeValidMethod, container, element, false);
    }

    public void removeAllAssociations(final NakedObject inObject) {
        try {
            if (clearMethod == null) {
                clear(getCollection(inObject));
                NakedObjectsContext.getObjectPersistor().objectChanged(inObject);
            } else {
                clearMethod.invoke(inObject.getObject(), (Object[]) null);
            }
        } catch (InvocationTargetException e) {
            invocationException("Exception executing " + clearMethod, e);
        } catch (IllegalAccessException ignore) {
            LOG.error("illegal access of " + clearMethod, ignore);
            throw new RuntimeException(ignore.getMessage());
        }
    }

    protected abstract void clear(Object collection);

    /**
     * Remove an associated object (the element) from the specified NakedObject in the association field
     * represented by this object.
     */
    public void removeAssociation(final NakedObject inObject, final NakedObject associate) {
        LOG.debug("local clear association " + associate + " from field " + getIdentifier() + " in " + inObject);

        Object element = associate.getObject();
        try {
            if (removeMethod == null) {
                removeElement(getCollection(inObject), element);
                NakedObjectsContext.getObjectPersistor().objectChanged(inObject);
            } else {
                removeMethod.invoke(inObject.getObject(), new Object[] { element });
            }
        } catch (IllegalArgumentException e) {
            throw new IllegalArgumentException(e.getMessage() + " object: " + inObject.getObject() + ", parameter: " + element);
        } catch (InvocationTargetException e) {
            invocationException("Exception executing " + removeMethod, e);
        } catch (IllegalAccessException ignore) {
            LOG.error("illegal access of " + removeMethod, ignore);
            throw new RuntimeException(ignore.getMessage());
        }
    }

    protected abstract void removeElement(Object collection, Object element);

    public String toString() {
        String methods = (getMethod == null ? "" : "GET") + (addMethod == null ? "" : " ADD")
                + (removeMethod == null ? "" : " REMOVE");

        return "OneToManyAssociation [name=\"" + getIdentifier() + "\", method=" + getMethod + ", methods=" + methods + ", type="
                + getSpecification() + " ]";
    }
}
// Copyright (c) Naked Objects Group Ltd.
