package org.nakedobjects.nof.reflect.java.reflect;

import java.util.Iterator;
import java.util.Set;
import java.util.SortedMap;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

import org.nakedobjects.noa.annotations.MemberOrderAnnotation;
import org.nakedobjects.nof.reflect.peer.MemberPeer;


/**
 * Represents a nested hierarchy of ordered members.
 * 
 * <p>
 * At each level the elements are either {@link MemberPeer}s or they are
 * instances of {@link OrderSet} represent a group of {@link MemberPeer}s that
 * have a {@link MemberOrderAnnotation} of the same name.
 * 
 * <p>
 * With no name, (ie <tt>name=""</tt> is the default), at the top level
 * <pre>
 * MemberOrder(sequence="1")
 * MemberOrder(sequence="1.1")
 * MemberOrder(sequence="1.2")
 * MemberOrder(sequence="1.2.1")
 * MemberOrder(sequence="1.3")
 * </pre>
 *
 * <p>
 * With names, creates a hierarchy.
 * <pre>
 * MemberOrder(sequence="1.1")                   // no parent
 * MemberOrder(sequence="1.2.1")
 * MemberOrder(sequence="1.3")
 * MemberOrder(name="abc", sequence="1")         // group is abc, parent is ""
 * MemberOrder(name="abc", sequence="1.2")
 * MemberOrder(name="abc,def", sequence="1")     // group is def, parent is abc
 * MemberOrder(name="abc,def", sequence="1.2")
 * </pre>
 *
 */
class DeweyOrderSet extends OrderSet {
    static DeweyOrderSet createOrderSet(final MemberPeer[] members) {
        
        SortedMap sortedMembersByGroup = new TreeMap();
        SortedSet nonAnnotatedGroup = new TreeSet(new MemberIdentifierComparator());
        
        // spin over all the members and put them into a Map of SortedSets
        // any non-annotated members go into additional nonAnnotatedGroup set.
        for(int i=0; i<members.length; i++) {
            final MemberPeer member = members[i];
            final MemberOrderAnnotation memberOrder = 
                (MemberOrderAnnotation) member.getExtension(MemberOrderAnnotation.class);
            if (memberOrder == null) {
                nonAnnotatedGroup.add(member);
                continue;
            }
            SortedSet sortedMembersForGroup = getSortedSet(sortedMembersByGroup, memberOrder.name());
            sortedMembersForGroup.add(member);
        }
        
        // add the non-annotated group to the first "" group.
        SortedSet defaultSet = getSortedSet(sortedMembersByGroup, "");
        defaultSet.addAll(nonAnnotatedGroup);

        
        // create OrderSets, wiring up parents and children.
        
        // since sortedMembersByGroup is a SortedMap, the 
        // iteration will be in alphabetical order (ie parent groups before their children). 
        Set groupNames = sortedMembersByGroup.keySet();
        SortedMap orderSetsByGroup = new TreeMap();
        
        for(Iterator iter = groupNames.iterator(); iter.hasNext(); ) {
            String groupName = (String)iter.next();
            DeweyOrderSet deweyOrderSet = new DeweyOrderSet(groupName);
            orderSetsByGroup.put(groupName, deweyOrderSet);
            ensureParentFor(orderSetsByGroup, deweyOrderSet);
        }

        // now populate the OrderSets

        for(Iterator iter = groupNames.iterator(); iter.hasNext(); ) {
            String groupName = (String)iter.next();
            DeweyOrderSet deweyOrderSet = (DeweyOrderSet) orderSetsByGroup.get(groupName);
            SortedSet sortedMembers = (SortedSet) sortedMembersByGroup.get(groupName);
            deweyOrderSet.addAll(sortedMembers);
            deweyOrderSet.copyOverChildren();
        }
        
        
        return (DeweyOrderSet) orderSetsByGroup.get("");
    }

    /**
     * Recursively creates parents all the way up to root (<tt>""</tt>),
     * along the way associating each child with its parent and adding
     * the child as an element of its parent.
     * 
     * @param orderSetsByGroup
     * @param deweyOrderSet
     */
    private static void ensureParentFor(
            SortedMap orderSetsByGroup,
            DeweyOrderSet deweyOrderSet) {
        String parentGroup = deweyOrderSet.getGroupPath();
        DeweyOrderSet parentOrderSet = 
            (DeweyOrderSet)orderSetsByGroup.get(parentGroup);
        if (parentOrderSet == null) {
            parentOrderSet = new DeweyOrderSet(parentGroup);
            orderSetsByGroup.put(parentGroup, parentOrderSet);
            if (!parentGroup.equals("")) {
                ensureParentFor(orderSetsByGroup, deweyOrderSet);
            }
        }
        // check in case at root
        if (deweyOrderSet != parentOrderSet) {
            deweyOrderSet.setParent(parentOrderSet);
            parentOrderSet.addChild(deweyOrderSet);
        }
    }

    /**
     * Gets the SortedSet with the specified group from the supplied Map of SortedSets.
     * 
     * <p>
     * If there is no such SortedSet, creates.
     * 
     * @param sortedMembersByGroup
     * @param groupName
     * @return
     */
    private static SortedSet getSortedSet(
            SortedMap sortedMembersByGroup,
            String groupName) {
        SortedSet sortedMembersForGroup;
        sortedMembersForGroup = (SortedSet) sortedMembersByGroup.get(groupName);
        if (sortedMembersForGroup == null) {
            sortedMembersForGroup = new TreeSet(new MemberOrderComparator(true));
            sortedMembersByGroup.put(groupName, sortedMembersForGroup);
        }
        return sortedMembersForGroup;
    }

    
    ///////////////////////////////////////////////////////////////////////////
    
    
    private DeweyOrderSet(String groupName) {
        super(groupName);
    }
    




    /**
     * Format is: <tt>abc,def:XXel/YYm/ZZch</tt>
     * <p>
     * where <tt>abc,def</tt> is group name, 
     * <tt>XX</tt> is number of elements, <tt>YY is number of members, and 
     * <tt>ZZ</tt> is number of child order sets.
     */
    public String toString() {
        return getGroupFullName() + ":" + size() + "el/" + (size()-childOrderSets.size()) + "m/" + childOrderSets.size() + "ch";
    }

}

// Copyright (c) Naked Objects Group Ltd.
