package org.nakedobjects.nof.reflect.java.reflect;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import org.apache.log4j.Logger;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.reflect.Consent;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.util.DebugString;
import org.nakedobjects.nof.reflect.peer.MemberIdentifier;
import org.nakedobjects.nof.reflect.peer.OneToOnePeer;
import org.nakedobjects.nof.reflect.peer.ReflectionException;


public class JavaOneToOneAssociation extends JavaField implements OneToOnePeer {
    private final static Logger LOG = Logger.getLogger(JavaOneToOneAssociation.class);
    private final Method associateMethod;
    private final Method dissociateMethod;
    private final Method validMethod;
 
    public JavaOneToOneAssociation(
            final MemberIdentifier identifier,
            final Class type,
            final DescriptiveMethods descriptiveMethods,
            final FieldMethods fieldMethods,
            final GeneralControlMethods controlMethods,
            final FieldFlags fieldFlags,
            final MemberSessionMethods sessionMethods) {
        super(identifier, type, descriptiveMethods, fieldMethods, controlMethods, fieldFlags, sessionMethods);
        this.associateMethod = remember(fieldMethods.getAssociateMethod());
        this.dissociateMethod = remember(fieldMethods.getDissociateMethod());
        this.validMethod = remember(controlMethods.getValidMethod1());
    }

    public void clearAssociation(final NakedObject inObject, final NakedObject associate) {
        LOG.debug("local clear association " + inObject + "/" + associate);

        try {
            if (dissociateMethod != null) {
                Object[] args;
                if (dissociateMethod.getParameterTypes().length == 0) {
                    args = new Object[]{};
                } else {
                    args = new Object[]{associate.getObject()};
                }
                dissociateMethod.invoke(inObject.getObject(), args);
            } else {
                setMethod.invoke(inObject.getObject(), new Object[] { null });
            }
        } catch (IllegalArgumentException e) {
            throw new IllegalArgumentException("set method expects a " + getSpecification().getFullName() + " object; not a "
                    + associate.getObject().getClass().getName());
        } catch (InvocationTargetException e) {
            invocationException("Exception executing " + setMethod, e);
        } catch (IllegalAccessException ignore) {
            LOG.error("illegal access of " + setMethod, ignore);
            throw new RuntimeException(ignore.getMessage());
        }
    }

    public void debugData(final DebugString debug) {
        super.debugData(debug);
        if (validMethod != null) {
            debug.appendln("Valid", validMethod);
        }
        if (associateMethod != null) {
            debug.appendln("Associate", associateMethod);
        }
        if (dissociateMethod != null) {
            debug.appendln("Dissociate", dissociateMethod);
        }
    }
    

    public NakedObject getDefault(NakedObject fromObject) {
        if (defaultMethod == null) {
            return null;
        }

        try {
            Object obj = defaultMethod.invoke(fromObject.getObject(), new Object[0]);
            if (obj == null) {
                return null;
            } else {
                return NakedObjectsContext.getObjectLoader().getAdapterForElseCreateAdapterForTransient(obj);
            }
        } catch (InvocationTargetException e) {
            invocationException("Exception executing " + defaultMethod, e);
            return null;
        } catch (IllegalAccessException ignore) {
            LOG.error("illegal access of " + defaultMethod, ignore);
            throw new ReflectionException(ignore);
        }
    }
    
    public NakedObject getAssociation(final NakedObject fromObject) {
        try {
            Object obj = getMethod.invoke(fromObject.getObject(), new Object[0]);

            if (obj == null) {
                return null;
            } else {
                NakedObject adapter;
                adapter = NakedObjectsContext.getObjectLoader().getAdapterForElseCreateAdapterForTransient(obj);
                return adapter;
            }

        } catch (InvocationTargetException e) {
            invocationException("Exception executing " + getMethod, e);
            return null;
        } catch (IllegalAccessException ignore) {
            LOG.error("illegal access of " + getMethod, ignore);
            throw new ReflectionException(ignore);
        }
    }

//    public Object getExtension(final Class cls) {
//        return null;
//    }
//
//    public Class[] getExtensions() {
//        return new Class[0];
//    }

    private boolean hasAddMethod() {
        return associateMethod != null;
    }

    /**
     * Set the data in an NakedObject. Passes in an existing object to for the EO to reference.
     */
    public void initAssociation(final NakedObject inObject, final NakedObject associate) {
        LOG.debug("local set association " + getIdentifier() + " in " + inObject + " with " + associate);

        try {
            setMethod.invoke(inObject.getObject(), new Object[] { associate == null ? null : associate.getObject() });
        } catch (IllegalArgumentException e) {
            throw new IllegalArgumentException(setMethod + " method doesn't expect a "
                    + associate.getObject().getClass().getName());
        } catch (InvocationTargetException e) {
            invocationException("Exception executing " + setMethod, e);
        } catch (IllegalAccessException ignore) {
            LOG.error("illegal access of " + setMethod, ignore);
            throw new RuntimeException(ignore.getMessage());
        }
    }

    public Consent isAssociationValid(final NakedObject target, final NakedObject reference) {
        return executeConsent(validMethod, target, reference, false);
    }

    public boolean isEmpty(final NakedObject fromObject) {
        try {
            Object obj = getMethod.invoke(fromObject.getObject(), new Object[0]);
            return obj == null;
        } catch (InvocationTargetException e) {
            invocationException("Exception executing " + getMethod, e);
            throw new ReflectionException(e);
        } catch (IllegalAccessException ignore) {
            LOG.error("illegal access of " + getMethod, ignore);
            throw new ReflectionException(ignore);
        }
    }
   
    public void setAssociation(final NakedObject inObject, final NakedObject associate) {
        LOG.debug("local set association " + getIdentifier() + " in " + inObject + " with " + associate);

        try {
            if (associate == null) {
                if (dissociateMethod != null) {
                    NakedObject association = getAssociation(inObject);
                    if (association != null) {
                        switch (dissociateMethod.getParameterTypes().length) {
                        case 0:
                            // clear method
                            if(association != null) {
                                dissociateMethod.invoke(inObject.getObject(), new Object[] {});
                            }
                            break;
                        case 1:
                            // old dissociate method
                            if(association != null) {
                                dissociateMethod.invoke(inObject.getObject(), new Object[] { association.getObject() });
                            }
                            break;
                        default:
                            // shouldn't happen
                            throw new RuntimeException("incorrect number of parameters for dissociate/clear method"); 
                        }
                    } else {
                        // shouldn't happen
                    }
                } else {
                    setMethod.invoke(inObject.getObject(), new Object[] { null });
                }
            } else {
                if (hasAddMethod()) {
                    associateMethod.invoke(inObject.getObject(), new Object[] { associate.getObject() });
                } else {
                    setMethod.invoke(inObject.getObject(), new Object[] { associate.getObject() });
                }
            }

        } catch (IllegalArgumentException e) {
            throw new IllegalArgumentException(setMethod + " method doesn't expect a " + associate.getObject().getClass().getName());
        } catch (InvocationTargetException e) {
            invocationException("Exception executing " + setMethod, e);
        } catch (IllegalAccessException ignore) {
            LOG.error("illegal access of " + setMethod, ignore);
            throw new RuntimeException(ignore.getMessage());
        }
    }

    public String toString() {
        String methods = (getMethod == null ? "" : "GET") + (setMethod == null ? "" : " SET")
                + (associateMethod == null ? "" : " ADD") + (dissociateMethod == null ? "" : " REMOVE");

        return "Association [name=\"" + getIdentifier() + "\", method=" + getMethod + ", methods="
                + methods + ", type=" + getSpecification() + " ]";
    }

}
// Copyright (c) Naked Objects Group Ltd.
