package org.nakedobjects.nof.reflect.java.reflect;

import java.util.Collection;
import java.util.Hashtable;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.nakedobjects.applib.value.Color;
import org.nakedobjects.applib.value.Date;
import org.nakedobjects.applib.value.DateTime;
import org.nakedobjects.applib.value.Image;
import org.nakedobjects.applib.value.Money;
import org.nakedobjects.applib.value.MultilineString;
import org.nakedobjects.applib.value.Password;
import org.nakedobjects.applib.value.Time;
import org.nakedobjects.applib.value.TimeStamp;
import org.nakedobjects.noa.adapter.NakedCollection;
import org.nakedobjects.noa.adapter.NakedValue;
import org.nakedobjects.noa.adapter.ObjectLoaderException;
import org.nakedobjects.noa.spec.NakedObjectSpecification;
import org.nakedobjects.nof.core.util.Assert;
import org.nakedobjects.nof.reflect.java.collection.ArrayAdapter;
import org.nakedobjects.nof.reflect.java.collection.CollectionCollectionAdapter;
import org.nakedobjects.nof.reflect.java.collection.ListCollectionAdapter;
import org.nakedobjects.nof.reflect.java.collection.SetCollectionAdapter;
import org.nakedobjects.nof.reflect.java.collection.VectorCollectionAdapter;
import org.nakedobjects.nof.reflect.java.value.ColorAdapter;
import org.nakedobjects.nof.reflect.java.value.DateAdapter;
import org.nakedobjects.nof.reflect.java.value.DateTimeAdapter;
import org.nakedobjects.nof.reflect.java.value.ImageAdapter;
import org.nakedobjects.nof.reflect.java.value.MoneyAdapter;
import org.nakedobjects.nof.reflect.java.value.MultilineStringAdapter;
import org.nakedobjects.nof.reflect.java.value.PasswordAdapter;
import org.nakedobjects.nof.reflect.java.value.TimeAdapter;
import org.nakedobjects.nof.reflect.java.value.TimeStampAdapter;
import org.nakedobjects.nof.reflect.spec.AbstractReflector;
import org.nakedobjects.nof.reflect.spec.AbstractSpecification;
import org.nakedobjects.nof.reflect.spec.ReflectionPeerBuilder;


public class JavaReflector extends AbstractReflector {
    private final static Logger LOG = Logger.getLogger(JavaReflector.class);
    private static final String PRIMITIVE_CLASSES = "Boolean Character Byte Double Float Integer Long Short Character String";
    private static final String NAKEDOBJECTS_VALUE_CLASS_PREFIX = "org.nakedobjects.applib.value.";
    private static final String NAKEDOBJECTS_FRAMEWORK_CLASS_PREFIX = "org.nakedobjects.nof.";
    private static final String NAKEDOBJECTS_ARCHITECTURE_CLASS_PREFIX = "org.nakedobjects.noa.";
    private static final String JAVA_CLASS_PREFIX = "java.";
    private static final String JAVA_LANG_PACKAGE_PREFIX = "java.lang.";

    private ClassStrategy classStrategy = new DefaultClassStrategy();
    private OneToManyStrategy[] oneToManyStrategies = new OneToManyStrategy[] { new ArrayStrategy(), new SetStrategy(),
            new JavaCollectionStrategy() };

    protected boolean collectionClass(Class cls, String className) {
        return java.util.Collection.class.isAssignableFrom(cls);
    }

    public ClassStrategy getClassStrategy() {
        return classStrategy;
    }

    public OneToManyStrategy[] getOneToManyStrategies() {
        return oneToManyStrategies;
    }

    protected boolean ignoreClass(Class cls, String className) {
        return className.startsWith(NAKEDOBJECTS_ARCHITECTURE_CLASS_PREFIX)
                || className.startsWith(NAKEDOBJECTS_FRAMEWORK_CLASS_PREFIX);
    }

    protected boolean ignoreJavaClass(Class cls, String className) {
        return className.startsWith(JAVA_CLASS_PREFIX) && cls != Object.class && cls != java.awt.Image.class
                && cls != java.sql.Date.class && cls != java.sql.Time.class && cls != java.math.BigDecimal.class
                && cls != java.math.BigInteger.class;
    }

    protected AbstractSpecification install(final Class cls, final ReflectionPeerBuilder builder) {
        return new JavaSpecification(cls, builder, this);
    }

    private boolean isJavaLangPrimitive(String className) {
        String shortClassName = className.substring(JAVA_LANG_PACKAGE_PREFIX.length());
        return PRIMITIVE_CLASSES.indexOf(shortClassName) >= 0;
    }

    public NakedObjectSpecification loadSpecification(Class cls) {
        return super.loadSpecification(classStrategy.getClass(cls));
    }

    protected boolean primitiveClass(Class cls, String className) {
        return className.startsWith(NAKEDOBJECTS_VALUE_CLASS_PREFIX)
                || (className.startsWith(JAVA_LANG_PACKAGE_PREFIX) && isJavaLangPrimitive(className))
                || java.util.Date.class.isAssignableFrom(cls) || java.awt.Image.class.isAssignableFrom(cls);
    }

    public void setClassStrategy(final ClassStrategy classStrategy) {
        this.classStrategy = classStrategy;
    }

    public void setOneToManyStrategies(OneToManyStrategy[] oneToManyStrategies) {
        this.oneToManyStrategies = oneToManyStrategies;
    }

    private final Hashtable adapterClasses = new Hashtable();

    private void addAdapterClass(Class valueClass, Class adapterClass) {
        NakedObjectSpecification valueSpecification = loadSpecification(valueClass);
        adapterClasses.put(valueSpecification, adapterClass);
    }

    /**
     * Creates adapter for Java primitives, or else delegates to the AdapterFactory class.
     */
    public NakedValue createValueAdapter(final NakedObjectSpecification specification) {
        NakedValue adapter;
        Class adapterClass = (Class) adapterClasses.get(specification);
        if (adapterClass != null) {
            try {
                adapter = (NakedValue) adapterClass.newInstance();
            } catch (InstantiationException e) {
                throw new ObjectLoaderException("Failed to instantiate the value adapter", e);
            } catch (IllegalAccessException e) {
                throw new ObjectLoaderException("Failed to create value adapter of type " + adapterClass
                        + "; could not access constructor", e);
            }
        } else if (specification.superclass() != null) {
            return createValueAdapter(specification.superclass());
        } else {
            throw new ObjectLoaderException("Failed to create value adapter class for " + specification.getFullName());
        }
        return adapter;
    }

    public NakedValue createValueAdapter(final Object object) {
        Assert.assertFalse("can't create an adapter for a NO Specification", object instanceof NakedObjectSpecification);

        // TODO this could be replaced with a hash to store value and adapter Classes, and a reflective call
        // to instantiate an instance of the adapter type given the value.
        if (object instanceof MultilineString) {
            return new MultilineStringAdapter((MultilineString) object);
        } else if (object instanceof Password) {
            return new PasswordAdapter((Password) object);
        } else if (object instanceof Color) {
            return new ColorAdapter((Color) object);
        } else if (object instanceof org.nakedobjects.applib.value.Date) {
            return new DateAdapter((org.nakedobjects.applib.value.Date) object);

        } else if (object instanceof java.sql.Date) { // must test before java.util.Date
            return new org.nakedobjects.nof.core.adapter.value.DateAdapter((java.sql.Date) object);
        } else if (object instanceof java.sql.Time) { // must test before java.util.Date
            return new org.nakedobjects.nof.core.adapter.value.TimeAdapter((java.sql.Time) object);
        } else if (object instanceof java.util.Date) {
            return new org.nakedobjects.nof.core.adapter.value.DateTimeAdapter((java.util.Date) object);

        } else if (object instanceof DateTime) {
            return new DateTimeAdapter((DateTime) object);
        } else if (object instanceof Time) {
            return new TimeAdapter((Time) object);
        } else if (object instanceof TimeStamp) {
            return new TimeStampAdapter((TimeStamp) object);
        } else if (object instanceof Image) {
            return new ImageAdapter((Image) object);
        } else if (object instanceof java.awt.Image) {
            return new org.nakedobjects.nof.core.adapter.value.ImageAdapter((java.awt.Image) object);
        } else if (object instanceof Money) {
            return new MoneyAdapter((Money) object);
        } else {
            throw new ObjectLoaderException("Failed to create value adapter for value " + object);
        }
    }

    // REVIEW this is creating an internal collection adapter; should the internal one-to-one assoc therefore
    // call the object loader?
    public NakedCollection createCollectionAdapter(final Object collection, final NakedObjectSpecification elementSpecification) {
        if (collection instanceof Vector) {
            return new VectorCollectionAdapter((Vector) collection, elementSpecification);
        } else if (collection instanceof Object[]) {
            return new ArrayAdapter((Object[]) collection, elementSpecification);
        } else if (collection instanceof List) {
            return new ListCollectionAdapter((List) collection, elementSpecification);
        } else if (collection instanceof Set) {
            return new SetCollectionAdapter((Set) collection, elementSpecification);
        } else if (collection instanceof Collection) {
            return new CollectionCollectionAdapter((Collection) collection, elementSpecification);
        } else {
            return null;
        }
    }

    protected void finalize() throws Throwable {
        super.finalize();
        LOG.info("finalizing reflector factory " + this);
    }

    public void init() {
        super.init();
        addAdapterClass(MultilineString.class, MultilineStringAdapter.class);
        addAdapterClass(Password.class, PasswordAdapter.class);
        addAdapterClass(Color.class, ColorAdapter.class);
        addAdapterClass(Date.class, DateAdapter.class);
        addAdapterClass(DateTime.class, DateTimeAdapter.class);
        addAdapterClass(Time.class, TimeAdapter.class);
        addAdapterClass(TimeStamp.class, TimeStampAdapter.class);
        addAdapterClass(Image.class, ImageAdapter.class);
        addAdapterClass(java.awt.Image.class, org.nakedobjects.nof.core.adapter.value.ImageAdapter.class);
        addAdapterClass(Money.class, MoneyAdapter.class);
    }

}
// Copyright (c) Naked Objects Group Ltd.
