package org.nakedobjects.nof.reflect.java.reflect;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.nakedobjects.noa.NakedObjectRuntimeException;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.adapter.NakedReference;
import org.nakedobjects.noa.adapter.Persistable;
import org.nakedobjects.noa.reflect.Consent;
import org.nakedobjects.noa.reflect.NakedObjectAction;
import org.nakedobjects.noa.reflect.NakedObjectField;
import org.nakedobjects.noa.reflect.NakedObjectMember;
import org.nakedobjects.noa.reflect.NakedObjectReflector;
import org.nakedobjects.noa.spec.Features;
import org.nakedobjects.noa.spec.NakedObjectSpecification;
import org.nakedobjects.noa.spec.NakedObjectSpecificationException;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.reflect.Allow;
import org.nakedobjects.nof.core.reflect.Veto;
import org.nakedobjects.nof.core.util.DebugInfo;
import org.nakedobjects.nof.core.util.DebugString;
import org.nakedobjects.nof.core.util.Dump;
import org.nakedobjects.nof.core.util.ToString;
import org.nakedobjects.nof.core.util.UnexpectedCallException;
import org.nakedobjects.nof.core.util.UnknownTypeException;
import org.nakedobjects.nof.reflect.peer.ObjectTitle;
import org.nakedobjects.nof.reflect.peer.OneToManyPeer;
import org.nakedobjects.nof.reflect.peer.OneToOnePeer;
import org.nakedobjects.nof.reflect.peer.ReflectionException;
import org.nakedobjects.nof.reflect.peer.ValuePeer;
import org.nakedobjects.nof.reflect.spec.AbstractSpecification;
import org.nakedobjects.nof.reflect.spec.ActionSet;
import org.nakedobjects.nof.reflect.spec.ReflectionPeerBuilder;

public class JavaSpecification extends AbstractSpecification implements DebugInfo, ExtensionHolderMutable {
    private class SubclassList {
        private Vector classes = new Vector();

        public void addSubclass(final NakedObjectSpecification subclass) {
            classes.addElement(subclass);
        }

        public boolean hasSubclasses() {
            return !classes.isEmpty();
        }

        public NakedObjectSpecification[] toArray() {
            NakedObjectSpecification[] classesArray = new NakedObjectSpecification[classes.size()];
            classes.copyInto(classesArray);
            return classesArray;
        }
    }

    private final static Logger LOG = Logger.getLogger(JavaSpecification.class);

    private static final Object[] NO_PARAMETERS = new Object[0];
    private NakedObjectAction[] classActions;
    private Method clearDirtyMethod;
    private String description;
    private Method iconMethod;
    private NakedObjectSpecification[] interfaces;
    private JavaIntrospector introspector;
    private Method isDirtyMethod;
    private Method isPersistableMethod;
    private Method markDirtyMethod;
    private Method lifecycleMethods[];
    private Persistable persistable;
    private String pluralName;
    private String shortName;
    private String singularName;
    private SubclassList subclasses;
    private ObjectTitle title;
    private NakedObjectField[] accessibleFields;
    private int type;
    private int features;
    private Class searchViaRepository;
    private boolean queryByExample;
    private Hashtable membersByMethod = new Hashtable();

    /**
     * @see #getExtension(Class).
     */
    private Class underlyingClass;


    public JavaSpecification(
            final Class cls,
            final ReflectionPeerBuilder builder,
            final JavaReflector specificationLoader) {
        introspector = new JavaIntrospector(cls, this, builder, specificationLoader, this);
        subclasses = new SubclassList();
        this.underlyingClass = cls;
        addExtension(underlyingClass);
        
        introspector.addAnnotationsAsExtensions();
    }

    public void addSubclass(final NakedObjectSpecification subclass) {
        subclasses.addSubclass(subclass);
    }

    /**
     * Can obtain the underlying class used to build this specification.
     * 
     * <p>
     * To do this, use (rather confusingly) <tt>Class.class</tt> as the key:
     * <pre>
     * Class underlyingClass = someJavaSpec.getExtension(Class.class);
     * </pre>
     */
    public Object getExtension(
            Class extensionKey) {
        return super.getExtension(extensionKey);
    }
    
    /**
     * As per {@link ExtensionHolderMutable}.
     */
    public void addExtension(
            Object extension) {
        super.addExtension(extension);
    }

    public void clearDirty(final NakedObject object) {
        if (clearDirtyMethod == null) {
            return;
        }

        try {
            clearDirtyMethod.invoke(object.getObject(), NO_PARAMETERS);
        } catch (IllegalArgumentException e) {
            throw new NakedObjectRuntimeException(e);
        } catch (IllegalAccessException e) {
            LOG.error("illegal access of " + clearDirtyMethod, e);
        } catch (InvocationTargetException e) {
            JavaMember.invocationException("Exception executing " + clearDirtyMethod, e);
        }
    }

    private NakedObjectField createNakedObjectField(final ReflectionPeerBuilder builder, final Object object) {
        NakedObjectField field;
        if (object instanceof ValuePeer) {
            field = builder.createField((ValuePeer) object);

        } else if (object instanceof OneToOnePeer) {
            field = builder.createField((OneToOnePeer) object);

        } else if (object instanceof OneToManyPeer) {
            field = builder.createField((OneToManyPeer) object);

        } else {
            throw new NakedObjectRuntimeException();
        }
        return field;
    }
    
    public void debugData(DebugString debug) {
        debug.blankLine();
        debug.appendln("Title", title);
        if (iconMethod != null) {
            debug.appendln("Icon", iconMethod);
        }
        if (isPersistableMethod != null) {
            debug.appendln("Persistable", isPersistableMethod);
        }
        debug.appendln("QBE", queryByExample);
        if (searchViaRepository != null) {
            debug.appendln("Search", searchViaRepository);
        }
        
        debug.blankLine();  
        debug.appendln("Life cycle methods");
        debug.indent();
        boolean none = true;
        for (int i = 0; i < lifecycleMethods.length; i++) {
            if (lifecycleMethods[i] != null) {
                debug.appendln(lifecycleMethods[i].toString());
                none = false;
            }
        }
        if (none) {
            debug.appendln("none");
        }
        debug.unindent();
        
        debug.blankLine();  
        debug.appendln("Extensions");
        debug.indent();
        none = true;
        Class[] extensionClasses = getExtensions();
        for (int i = 0; i < extensionClasses.length; i++) {
            debug.appendln(getExtension(extensionClasses[i]).toString());
            none = false;
        }
        if (none) {
            debug.appendln("none");
        }
        debug.unindent();
        
        

    }

    public String debugTitle() {
        return "NO Member Specification";
    }
    
    public NakedObjectField[] getStaticallyVisibleFields() {
        if (accessibleFields == null) {
            NakedObjectField[] selectedFields = new NakedObjectField[fields.length];
            int v = 0;
            for (int i = 0; i < fields.length; i++) {
                boolean isAccessible = fields[i].isVisible();
                if (isAccessible) {
                    selectedFields[v++] = fields[i];
                }
            }

            accessibleFields = new NakedObjectField[v];
            System.arraycopy(selectedFields, 0, accessibleFields, 0, v);
        }
        return accessibleFields;
    }

    private NakedObjectAction getAction(
            final NakedObjectAction[] availableActions,
            final NakedObjectAction.Type type,
            final String id,
            final NakedObjectSpecification[] parameters) {
        outer: for (int i = 0; i < availableActions.length; i++) {
            NakedObjectAction action = availableActions[i];
            if (action.getActions().length > 0) {
                NakedObjectAction a = getAction(action.getActions(), type, id, parameters);
                if(a != null) {
                    return a;
                }
            } else {
                if (action.getType().equals(type)) {
                    if (id == null || id.equals(action.getId())) {
                        if (action.getParameterTypes().length == parameters.length) {
                            for (int j = 0; j < parameters.length; j++) {
                                if (!parameters[j].isOfType(action.getParameterTypes()[j])) {
                                    continue outer;
                                }
                            }
                            return action;
                        }
                    }
                }
            }
        }
        return null;
    }

    protected NakedObjectAction[] getActions(
            final NakedObjectAction[] availableActions,
            final NakedObjectAction.Type type) {
        Vector actions = new Vector();
        for (int i = 0; i < availableActions.length; i++) {
            NakedObjectAction action = availableActions[i];
            NakedObjectAction.Type actionType = action.getType();
            if (actionType == NakedObjectAction.SET) {
                NakedObjectAction[] subActions = ((ActionSet) action).getActions();
                for (int j = 0; j < subActions.length; j++) {
                    if (subActions[i].getType().equals(type)) {
                        actions.addElement(action);                        
                        break;
                    }
                }
            } else if (actionType.equals(type)) {
                actions.addElement(action);
            }
        }

        NakedObjectAction[] results = new NakedObjectAction[actions.size()];
        actions.copyInto(results);
        return results;
    }

    public NakedObject getAggregate(NakedObject object) {
        if (! Features.isAggregated(this)) {
            throw new UnexpectedCallException("No aggregate method");
        }
        return null; 
        /*
        Object aggregate = ((Aggregated) object).getAggregate();
        return NakedObjectsContext.getObjectLoader().getAdapterForElseCreateAdapterForTransient(aggregate);
        */
    }

    public NakedObjectAction getClassAction(
            final NakedObjectAction.Type type,
            final String id,
            final NakedObjectSpecification[] parameters) {
            // TODO also need to find service actions as well, see getClassActions
            return getAction(classActions, type, id, parameters);
    }

    public String getDescription() {
        return description == null ? "" : description;
    }

    public int getFeatures() {
        return features;
    }

    public NakedObjectField getField(final String id) {
        // TODO put fields into hash
        for (int i = 0; i < fields.length; i++) {
            if (fields[i].getId().equals(id)) {
                return fields[i];
            }
        }
        throw new NakedObjectSpecificationException("No field called '" + id + "' in '" + getSingularName() + "'");
    }

    public String getIconName(NakedReference reference) {
        if (iconMethod == null) {
            return null;
        }
        try {
            return (String) iconMethod.invoke(reference.getObject(), NO_PARAMETERS);
        } catch (IllegalArgumentException e) {
            throw new NakedObjectRuntimeException(e);
        } catch (IllegalAccessException e) {
            LOG.error("illegal access of " + isDirtyMethod, e);
            return null;
        } catch (InvocationTargetException e) {
            JavaMember.invocationException("Exception executing " + isDirtyMethod, e);
            return null;
        }
    }

    public NakedObjectAction getObjectAction(
            final NakedObjectAction.Type type,
            final String id,
            final NakedObjectSpecification[] parameters) {
        // TODO also need to find service actions as well, see getObjectActions
        return getAction(objectActions, type, id, parameters);
    }

    public String getPluralName() {
        return pluralName;
    }

    public Class getSearchViaRepository() {
        return searchViaRepository;
    }
    
    public String getShortName() {
        return shortName;
    }

    public String getSingularName() {
        return singularName;
    }

    public String getTitle(final NakedObject object) {
        String titleString = title.title(object);
        if(titleString == null || titleString.equals("")) {
            return (Features.isService(this) ? "" : "Untitled ") + getSingularName();
        } else {
            return titleString;
        }
    }

    public int getType() {
        return type;
    }
    
    public NakedObjectField[] getDynamicallyVisibleFields(final NakedObject object) {
        getStaticallyVisibleFields();
        NakedObjectField[] selectedFields = new NakedObjectField[accessibleFields.length];
        int v = 0;
        for (int i = 0; i < accessibleFields.length; i++) {
            if (accessibleFields[i].isVisible() && accessibleFields[i].isVisible(object)) {
                selectedFields[v++] = accessibleFields[i];
            }
        }

        NakedObjectField[] visibleFields = new NakedObjectField[v];
        System.arraycopy(selectedFields, 0, visibleFields, 0, v);
        return visibleFields;
    }

    public boolean hasSubclasses() {
        return subclasses.hasSubclasses();
    }

    public NakedObjectSpecification[] interfaces() {
        return interfaces;
    }
    
    public void introspect(ReflectionPeerBuilder builder) {
        if (introspector == null) {
            throw new ReflectionException("Introspection already taken place, cannot introspect again");
        }

        introspector.introspect();

        clearDirtyMethod = introspector.getClearDirtyMethod();
        isDirtyMethod = introspector.getIsDirtyMethod();
        markDirtyMethod = introspector.getMarkDirtyMethod();

         
        iconMethod = introspector.getIconMethod();
        
        OrderSet orderedActions =  introspector.getClassActions();
        classActions = orderActions(builder, orderedActions);
       
        OrderSet orderedFields = introspector.getFields();
        if(orderedFields != null) {
            fields = orderFields(builder, orderedFields);
        }


        orderedActions = introspector.getObjectActions();
        objectActions = orderActions(builder, orderedActions);

        className = introspector.getFullName();
        description = introspector.getDescription();
        pluralName = introspector.pluralName();
        singularName = introspector.singularName();
        shortName = introspector.shortName();

        title = introspector.title();

        features += introspector.isAbstract() ? NakedObjectSpecification.ABSTRACT : 0;
        features += introspector.isAggregated() ? NakedObjectSpecification.AGGREGATED : 0;
        features += introspector.isEntity() ? NakedObjectSpecification.ENTITY : 0;
        features += introspector.isExtendable() ? NakedObjectSpecification.EXTENDABLE : 0;
        features += introspector.isBounded() ? NakedObjectSpecification.BOUNDED : 0;
        features += introspector.isSearchable() ? NakedObjectSpecification.SEARCHABLE : 0;
        features += introspector.isImmutableOncePersisted() ? NakedObjectSpecification.IMMUTABLE_ONCE_PERSISTED : 0;
        features += introspector.isAlwaysImmutable() ? NakedObjectSpecification.ALWAYS_IMMUTABLE : 0;
        features += introspector.isCached() ? NakedObjectSpecification.CACHED : 0;
              
        type = introspector.getType();

        lifecycleMethods = introspector.lifecycleMethods();
        isPersistableMethod = introspector.isPersistableMethod();
        
        persistable = introspector.persistable();

        String superclassName = introspector.getSuperclass();
        String[] interfaceNames = introspector.getInterfaces();

        NakedObjectReflector loader = NakedObjectsContext.getReflector();
        if (superclassName != null) {
            superClassSpecification = loader.loadSpecification(superclassName);
            if (superClassSpecification != null) {
                LOG.debug("  Superclass " + superclassName);
                // superclass.subclasses.addSubclass(this);
                superClassSpecification.addSubclass(this);
            }
        }

        interfaces = new NakedObjectSpecification[interfaceNames.length];
        for (int i = 0; i < interfaceNames.length; i++) {
            interfaces[i] = loader.loadSpecification(interfaceNames[i]);
            // interfaces[i].subclasses.addSubclass(this);
            interfaces[i].addSubclass(this);
        }

        queryByExample = introspector.getQueryByExample();
        searchViaRepository = introspector.getSearchViaRepository();
        
        introspector = null;
    }

    public boolean isDirty(final NakedObject object) {
        if (isDirtyMethod == null) {
            return false;
        }

        try {
            Boolean isDirty = (Boolean) isDirtyMethod.invoke(object.getObject(), NO_PARAMETERS);
            return isDirty.booleanValue();
        } catch (IllegalArgumentException e) {
            throw new NakedObjectRuntimeException(e);
        } catch (IllegalAccessException e) {
            LOG.error("illegal access of " + isDirtyMethod, e);
            return false;
        } catch (InvocationTargetException e) {
            JavaMember.invocationException("Exception executing " + isDirtyMethod, e);
            return false;
        }
    }

    /**
     * Determines if this class represents the same class, or a subclass, of the specified class.
     */
    public boolean isOfType(final NakedObjectSpecification specification) {
        if (specification == this) {
            return true;
        } else {
            if (interfaces != null) {
                for (int i = 0, len = interfaces.length; i < len; i++) {
                    if (interfaces[i].isOfType(specification)) {
                        return true;
                    }
                }
            }
            if (superClassSpecification != null) {
                return superClassSpecification.isOfType(specification);
            }
        }
        return false;
    }
    
    public Consent isPersistable(NakedObject object) {
        if(isPersistableMethod == null) {
            return Allow.DEFAULT;
        } else {
            try {
                String reason = (String) isPersistableMethod.invoke(object.getObject(), NO_PARAMETERS);
                return Veto.create(reason == null, "", reason);
            } catch (IllegalArgumentException e) {
                throw new NakedObjectRuntimeException(e);
            } catch (IllegalAccessException e) {
                LOG.error("illegal access of " + isPersistableMethod, e);
                return Allow.DEFAULT;
            } catch (InvocationTargetException e) {
                JavaMember.invocationException("Exception executing " + isPersistableMethod, e);
                return null;
            }

        }
    }

    public void lifecycleEvent(final NakedObject object, final int event) {
        Method method = lifecycleMethods[event];
        if(method != null) {
            try {
                method .invoke(object.getObject(), NO_PARAMETERS);
            } catch (IllegalArgumentException e) {
                throw new NakedObjectRuntimeException(e);
            } catch (IllegalAccessException e) {
                LOG.error("illegal access of " + method, e);
            } catch (InvocationTargetException e) {
                JavaMember.invocationException("Exception executing " + method, e);
            }
        }
    }
    
    public void markAsService() {
        NakedObjectField[] fields = getFields();
        if (fields != null && fields.length > 0) {
            String fieldNames = "";
            for (int i = 0; i < fields.length; i++) {
                String name = fields[i].getName();
                if ("Id".indexOf(name) == -1) {
                    fieldNames += (fieldNames.length() > 0 ? ", " : "") + name;
                }
            }
            if (fieldNames.length() > 0) {
                throw new NakedObjectSpecificationException("Service object " + getFullName()
                        + " should have no fields, but has: " + fieldNames);
            }
        }
        /* TODO replace when specification knows of its hidden methods
        if (objectActions != null && objectActions.length == 0) {
            throw new NakedObjectSpecificationException("Service object " + getFullName()
                    + " should have at least one user action");
        }
        */
        features += NakedObjectSpecification.SERVICE;
    }

    public void markDirty(final NakedObject object) {
        if (markDirtyMethod == null) {
            return;
        }

        try {
            markDirtyMethod.invoke(object.getObject(), NO_PARAMETERS);
        } catch (IllegalArgumentException e) {
            throw new NakedObjectRuntimeException(e);
        } catch (IllegalAccessException e) {
            LOG.error("illegal access of " + isDirtyMethod, e);
        } catch (InvocationTargetException e) {
            JavaMember.invocationException("Exception executing " + isDirtyMethod, e);
        }
    }

    private NakedObjectAction[] orderActions(final ReflectionPeerBuilder builder, final OrderSet order) {
        NakedObjectAction[] actions = new NakedObjectAction[order.size()];
        Enumeration elements = order.elements();
        int actionCnt = 0;
        while (elements.hasMoreElements()) {
            Object element = elements.nextElement();
            if (element instanceof JavaAction) {
                JavaAction javaAction = (JavaAction) element;
                NakedObjectAction nakedObjectAction = builder.createAction(javaAction);
                actions[actionCnt++] = nakedObjectAction;
                updateMembersByMethod(javaAction, nakedObjectAction);
            } else if (element instanceof OrderSet) {
                OrderSet set = ((OrderSet) element);
                actions[actionCnt++] = new ActionSet("", set.getGroupFullName(), orderActions(builder, set));
            } else {
                throw new UnknownTypeException(element);
            }
        }

        if (actionCnt < actions.length) {
            NakedObjectAction[] actualActions = new NakedObjectAction[actionCnt];
            System.arraycopy(actions, 0, actualActions, 0, actionCnt);
            return actualActions;
        }
        return actions;
    }

    private NakedObjectField[] orderFields(final ReflectionPeerBuilder builder, final OrderSet order) {
        NakedObjectField[] fields = new NakedObjectField[order.size()];
        Enumeration elements = order.elements();
        int actionCnt = 0;
        while (elements.hasMoreElements()) {
            Object element = elements.nextElement();
            if (element instanceof JavaField) {
                JavaField javaField = (JavaField)element;
                NakedObjectField nakedObjectField = createNakedObjectField(builder, javaField);
                fields[actionCnt++] = nakedObjectField;
                updateMembersByMethod(javaField, nakedObjectField);
            } else if (element instanceof OrderSet) {
                // Not supported at present
            } else {
                throw new UnknownTypeException(element);
            }
        }

        if (actionCnt < fields.length) {
            NakedObjectField[] actualActions = new NakedObjectField[actionCnt];
            System.arraycopy(fields, 0, actualActions, 0, actionCnt);
            return actualActions;
        }
        return fields;
    }

    private void updateMembersByMethod(
            JavaMember javaMember, NakedObjectMember nakedObjectMember) {
        Method[] methods = javaMember.getMethods();
        for(int i=0; i<methods.length; i++) {
            this.membersByMethod.put(methods[i], nakedObjectMember);
        }
    }


    public Persistable persistable() {
        return persistable;
    }

    public boolean queryByExample() {
        return queryByExample;
    }
    
    public NakedObjectSpecification[] subclasses() {
        return subclasses.toArray();
    }
    
    public String toString() {
        ToString str = new ToString(this);
        str.append("class", className);
        str.append("type", Dump.typeNameFor(this));
        str.append("persistable", persistable);
        str.append("superclass", superClassSpecification == null ? "Object" : superClassSpecification.getFullName());
        return str.toString();
    }

    
    public NakedObjectMember getMember(Method method) {
        Object memberObj = membersByMethod.get(method);
        return (NakedObjectMember) memberObj;
    }



}
// Copyright (c) Naked Objects Group Ltd.
